/*******************************************************************************
* Freescale Semiconductor Inc.
* (c) Copyright 2014 Freescale Semiconductor, Inc.
* ALL RIGHTS RESERVED.
********************************************************************************
Services performed by FREESCALE in this matter are performed AS IS and without
any warranty. CUSTOMER retains the final decision relative to the total design
and functionality of the end product. FREESCALE neither guarantees nor will be
held liable by CUSTOMER for the success of this project.
FREESCALE DISCLAIMS ALL WARRANTIES, EXPRESSED, IMPLIED OR STATUTORY INCLUDING,
BUT NOT LIMITED TO, IMPLIED WARRANTY OF MERCHANTABILITY OR FITNESS FOR
A PARTICULAR PURPOSE ON ANY HARDWARE, SOFTWARE ORE ADVISE SUPPLIED 
TO THE PROJECT BY FREESCALE, AND OR NAY PRODUCT RESULTING FROM FREESCALE 
SERVICES. IN NO EVENT SHALL FREESCALE BE LIABLE FOR INCIDENTAL OR CONSEQUENTIAL 
DAMAGES ARISING OUT OF THIS AGREEMENT.
CUSTOMER agrees to hold FREESCALE harmless against any and all claims demands 
or actions by anyone on account of any damage, or injury, whether commercial,
contractual, or tortuous, rising directly or indirectly as a result 
of the advise or assistance supplied CUSTOMER in connection with product, 
services or goods supplied under this Agreement.
********************************************************************************
* File:             main.c
* Owner:            b05111
* Version:          1.1
* Date:             Jul-29-2015
* Classification:   General Business Information
* Brief:            pin toggle stationery for MPC5675KEVB
********************************************************************************
* Detailed Description: 
* Application performs basic initialization, setup PLL to maximum allowed freq.,
* initializes interrupts, blinking one LED by interrupt, second LED by software
* loop, initializes and display notice via UART terminal and then terminal ECHO.
* The example configures the device for maximum performance (OPTIMIZATIONS_ON)
* by initialization of instruction/data cache and enabling of branch prediction.
* Example suppose MCU is configured for LSM (Lock-step mode). 
* Its intention is to offer advanced startup code additional to CW stationery.
*
* ------------------------------------------------------------------------------
* Test HW:        MPC5675KEVB
* MCU:            PPC5675KFMMSJ in Lock-Step mode
* Fsys:           180/150 MHz CORE_CLK
* Debugger:       Lauterbach Trace32
*                 PeMicro USB-ML-PPCNEXUS
* Target:         RAM, internal_FLASH
* Terminal:       19200-8-no parity-1 stop bit-no flow control on eSCI_A
* EVB connection: default 
*
********************************************************************************
Revision History:
0.1       Aug-16-2011   b08110  Initial Version
0.2       Sep-02-2011   b08110  Fixed PLL0 freq and CLKOUT freq settings
0.3       Sep-03-2011   b08110  PIT3 1s interrupt to toggle LED1.
0.4       Oct-19-2011   b08110  Added clear NCF 12 (SAFE mode after reset
                                due to NCF 12 timeout. STCU shadow flash
                                instructions per AN4322.)
0.5       Feb-16-2012   b08110  Fixed peri set1 frequency.
                                per RM, max for this clock is 45 MHz.
                                thus, we must set MC_CGM.SC_DC1 = 0x81;
                                - divide system clock (90 MHz) by 2
0.6       Mar-07-2012   b08110  Mode Entry chnage of the system clock source
                                from IRC16 to PLL0 done with maximal ODF.
                                After change, the ODF is stepped to reach target
                                system clock frequency (ODF can be changed
                                on MPC5675K while running system clock from it).
                                This is to prevent VDD1.2 lowvoltage destructive
                                reset on the EVB at the time when the system
                                clock is switched from 16M IRC to PLL0.
0.7       Jul-01-2014   b05111  Added printf functionality
0.8_LSM   Nov-10-2014   b05111  ModesAndClks_init revised, LSM and DPM versions
0.9_LSM   Nov-14-2014   b05111  Added Optimizations, cmm scripts changed
1.0_LSM   Mar-27-2015   b05111  INIT_Derivative for RAM_target, XBAR+CU init
1.1_LSM   Jul-29-2015   b05111  Corrected Optimizations.c
*******************************************************************************/

#include <stdio.h>
#include "MPC5675K.h"
#include "Exceptions.h"
#include "IntcInterrupts.h"
#include "uart.h"
#include "clock.h"
#include "Optimizations.h"

/*******************************************************************************
* External objects
*******************************************************************************/

/*******************************************************************************
* Global variables
*******************************************************************************/

/*******************************************************************************
* Constants and macros
*******************************************************************************/

/* Enable caches, branch prediction and optimize flash setting */
#define OPTIMIZATIONS_ON 1

#define FCCU_NCFK_KEY 0xAB3498FE
#define FCCU_CFK_KEY  0x618B7A50

/* PCR numbers */
#define FLEXPWM0_B0 59
#define FLEXPWM0_B1 62

#define LED0_pad FLEXPWM0_B0
#define LED1_pad FLEXPWM0_B1

/*******************************************************************************
* Local types
*******************************************************************************/

/*******************************************************************************
* Local function prototypes
*******************************************************************************/
static void HW_init(void);
static void SysClk_init(void);
static void ModesAndClks_init(void);
static void PeriClkGen_init(void);
void DisableWatchdog(void);
static void ClearNCF(void);
static void ClearCF(void);
static void ClearFails(void);
static void GPIO_init(void);
static void Delay(void);
static void PIT3_init(void);
static void PIT3_isr(void);
static void Terminal_ECHO(void);


/*******************************************************************************
Function Name : HW_init
Engineer      : b05111
Date          : Nov-10-2014
Parameters    : 
Modifies      : 
Returns       : 
Notes         : initialization of the hw for the purposes of this example
Issues        : 
*******************************************************************************/
static void HW_init(void)
{
    DisableWatchdog();
    ClearFails();    
    SysClk_init();    
    GPIO_init();
    
    #if OPTIMIZATIONS_ON
        Global_optimizations();   
        Core_optimizations();
    #endif 
}

/*******************************************************************************
Function Name : SysClk_init
Engineer      : b05111
Date          : Nov-10-2014
Parameters    : 
Modifies      : 
Returns       : 
Notes         : 
Issues        : 
*******************************************************************************/
static void SysClk_init(void)
{
    ModesAndClks_init();
    PeriClkGen_init();
}

/*******************************************************************************
Function Name : ModesAndClks_init
Engineer      : b05111
Date          : Nov-10-2014
Parameters    : 
Modifies      : 
Returns       : 
Notes         : - DRUN->RUN1->RUN0->RUN1->RUN1
                - enable CLKOUT
Issues        : Progressive clock switching can be too fast for certain
                regulators - in cases issues manual slow transition can be 
                implemented as recommended at the end of FMPLL chapter in RM  
*******************************************************************************/
static void ModesAndClks_init(void) 
{
    /* CLOCK ARCHITECTURE EXCERPT:

    2 MCU variants: a)CORE_CLK up to 180MHz (SYS_CLK up to 90)
                    b)CORE_CLK up to 150MHz (SYS_CLK up to 75)

    Clocking scheme based on NVUSRO non-volatile register (mirror in SSCM_UOPS):

    1) NVUSRO[CR] = 0b11 (default) - 1:2 Mode  fixed divider  additional divider
    =>      DRAMC_CLK    up to 180/150 MHz       1              none
            CORE_CLK     up to 180/150           1              none
            SYS_CLK      up to  90/75            2              none
            FLASH_CLK    up to  45/37.5          2              none
            PERI0_CLK    up to  90/75            1              CGM_SC_DC0
            PERI1_CLK    up to  45/37.5          2              CGM_SC_DC1

    2) NVUSRO[CR] = 0b10 - 1:3 Mode 
    =>      DRAMC_CLK    up to 180/150 MHz       1              none
            CORE_CLK     up to 180/150           1              none
            SYS_CLK      up to  60/50            3              none
            FLASH_CLK    up to  30/25            2              none
            PERI0_CLK    up to  60/50            1              CGM_SC_DC0
            PERI1_CLK    up to  30/25            2              CGM_SC_DC1
      
    3) NVUSRO[CR] = 0b01 - 1:1 Mode 
    =>      DRAMC_CLK    up to 180/150 MHz       1              none
            CORE_CLK     up to  90/75            2              none
            SYS_CLK      up to  90/75            2              none
            FLASH_CLK    up to  45/37.5          2              none
            PERI0_CLK    up to  90/75            1              CGM_SC_DC0
            PERI1_CLK    up to  45/37.5          2              CGM_SC_DC1

    General constraints:
    XOSC_CLK = 4-40 MHz, fvco = 256-512 MHz, DRAMC_CLK = 60-180 MHz, 
    CORE_CLK = 8-180 MHz, SYS_CLK = 8-90 MHz, FLASH_CLK = 4-45 MHz, 
    FMPLL0_CLK = 16-180 MHz, FMPLL1_1D0_CLK = 16-120 MHz, CLK_OUT = 2-60 MHz,
    FMPLL1_1D1_CLK = 42.66-80 MHz, PERI0_CLK = 8-90 MHz, PERI1_CLK = 4-45 MHz,
    MOTC_CLK = 4-120 MHz, FR_CLK = 40 or 80 MHz, CANPE_CLK = 4-40 MHz.
    
    Preconditions for code below:
    - XOSC = nounted 40MHz crystal (jumpers J42 and J43 in position 1-2)
    - Suspected MCU variant is: a)CORE_CLK up to 180MHz (SYS_CLK up to 90)
    - Suspected mode is: 1) NVUSRO[CR] = 0b11 (default) - 1:2 Mode
    - FMPLL0_CLK choosen as System Clock Selector Input Clock */

    int32_t cnt = 0;
    
    /* Enable DRUN, RUN0, RUN1 SAFE, RESET modes */
    ME.MER.R = 0x0000003D;
    
    /* Mode Transition to enter RUN1 mode:    
       RUN1 cfg: 16MHzIRCON,OSC0ON,PLL0OFF,PLL1OFF,syclk=16MIRC */
    ME.RUN[1].R = 0x001F0030;
    ME.MCTL.R = 0x50005AF0;             // Enter RUN1 Mode & Key       
    ME.MCTL.R = 0x5000A50F;             // Enter RUN1 Mode & Inverted Key
    while(0 == ME.GS.B.S_XOSC) {};      // Wait for mode entry to complete
    while(1 == ME.GS.B.S_MTRANS) {}     // Wait for mode transition to complete
    while(5 != ME.GS.B.S_CURRENTMODE) {};   // Check RUN1 mode has been entered 
    
 
    /* Initialize PLL before turning it on:
       FMPLL0_CLK = XOSC_CLK*NDIV/IDF/ODF
       fvco must be from 256 MHz to 512 MHz
       we want fsys = 180 MHz
       fvco = FMPLL0_CLK * ODF = 180 MHz * 2 = 360 MHz => OK
       FMPLL0_CLK =  40*36/4/2 = 180 MHz
       FMPLL0_CLK choosen as System Clock Selector Input Clock */

    CGM.AC3_SC.R = 0x01000000;  // Select Xosc as PLL0 source clock
    CGM.AC4_SC.R = 0x01000000;  // Select Xosc as PLL1 source clock

#if defined(CORE_CLOCK_180MHz)
    /* FMPLL0_CLK = 180 MHz */
    CGM.FMPLL[0].CR.B.IDF = 0x3;  // FMPLL0 IDF=3 --> divide by 4
    CGM.FMPLL[0].CR.B.ODF = 0x0;  // FMPLL0 ODF=0 --> divide by 2
    CGM.FMPLL[0].CR.B.NDIV = 36;  // FMPLL0 NDIV=36 --> divide by 36

#elif defined(CORE_CLOCK_150MHz)
    /* FMPLL0_CLK = 150 MHz */
    CGM.FMPLL[0].CR.B.IDF = 0x5;  // FMPLL0 IDF=5 --> divide by 6
    CGM.FMPLL[0].CR.B.ODF = 0x0;  // FMPLL0 ODF=0 --> divide by 2
    CGM.FMPLL[0].CR.B.NDIV = 45;  // FMPLL0 NDIV=45 --> divide by 45
								
#else // error
    while (1) { asm(nop); }
#endif
    
    /* FMPLL1_1D0_CLK = 120 MHz */
    CGM.FMPLL[1].CR.B.IDF = 0x5;  // FMPLL0 IDF=5 --> divide by 8
    CGM.FMPLL[1].CR.B.ODF = 0x1;  // FMPLL0 ODF=1 --> divide by 4
    CGM.FMPLL[1].CR.B.NDIV = 72;  // FMPLL0 NDIV=72 --> divide by 64

    /* enable progressive clock switching */ 
    /* Note: PHI_PCS output only on PLL_0 */
    CGM.FMPLL[0].CR.B.EN_PLL_SW = 1; 	
        
    ME.RUNPC[0].R = 0x000000FE;  // peripheral run in all modes
    ME.RUNPC[1].R = 0x00000000;  // peripheral frozen in all modes
    ME.LPPC[0].R = 0x00000000;   // disable peripheral run in LP modes
        
    /* disable all clocks that ca be disabled */
    CGM.SC_DC0.R = 0x00000000;   // disable peripheral set 0/1 clocks
        // note: header file defines this register different way than RM
        // SC_DC0 compounds here from SC_DC0 + SC_DC1 + 16 zero padding
    CGM.AC0_DC0.R = 0x00000000;  // disable Motor control clock
    CGM.AC1_DC0.R = 0x00000000;  // disable FlexRay clock
    CGM.AC2_DC0.R = 0x00000000;  // disable FlexCAN clock

    /* Mode Transition to enter RUN0 mode:
       RUN0 cfg: 16MHzIRCON,OSC0ON,PLL0ON,PLL1off,syclk=16M IRC */
    ME.RUN[0].R = 0x001F0070;           
    ME.MCTL.R = 0x40005AF0;             // Enter RUN0 Mode & Key
    ME.MCTL.R = 0x4000A50F;             // Enter RUN0 Mode & Inverted Key    
    while (1 == ME.GS.B.S_MTRANS) {}    // Wait for mode transition to complete
    while(4 != ME.GS.B.S_CURRENTMODE) {};   // Check RUN0 mode has been entered
    
    /* 1 SRAM waistate (default) */
    ECSM_0.MUDCR.R = 0x40000000; 
    
    /* max divider when switching FMPLL0_CLK from IRC to PLL0 */
    CGM.FMPLL[0].CR.B.ODF = 0x3;   // FMPLL0 ODF=3 --> divide by 16  
    
    /* Mode Transition to enter RUN1 mode:
       RUN1 cfg: 16MHzIRCON,OSC0ON,PLL0ON,PLL1off,syclk=PLL0 */
    ME.RUN[1].R = 0x001F0074;           
    ME.MCTL.R = 0x50005AF0;             // Enter RUN1 Mode & Key
    ME.MCTL.R = 0x5000A50F;             // Enter RUN1 Mode & Inverted Key
    while(1 == ME.GS.B.S_MTRANS) {}     // Wait for mode transition to complete
    while(5 != ME.GS.B.S_CURRENTMODE) {};   // Check RUN1 mode has been entered
    
    CGM.FMPLL[0].CR.B.ODF = 0x2;   // FMPLL0 ODF=2 --> divide by 8
    CGM.FMPLL[0].CR.B.ODF = 0x1;   // FMPLL0 ODF=1 --> divide by 4
    
    /* delay for the VDD reg to stabilize ....
       otherwise on the EVB, low voltage 1.2VDD destructive reset might occur */   
    for(cnt=0;cnt<300;cnt++) {} 
    CGM.FMPLL[0].CR.B.ODF = 0x0;   // FMPLL0 ODF=0 --> divide by 2

    /* Mode Transition to enter RUN1 mode:   
       RUN1 cfg: 16MHzIRCON,OSC0ON,PLL0ON,PLL1ON,syclk=PLL0 */ 
    ME.RUN[1].R = 0x001F00F4;      
    ME.MCTL.R = 0x50005AF0;             // Enter RUN1 Mode & Key
    ME.MCTL.R = 0x5000A50F;             // Enter RUN1 Mode & Inverted Key
    while(1 == ME.GS.B.S_MTRANS) {}     // Wait for mode transition to complete
    while(5 != ME.GS.B.S_CURRENTMODE) {};   // Check RUN1 mode has been entered
    
    /* peri set 0 divide by 1, per set1 divide by 2 */    
    CGM.SC_DC0.R = 0x80810000;
        // note: header file defines this register different way than RM
        // SC_DC0 compounds here from SC_DC0 + SC_DC1 + 16 zero padding
    
    /* set CLKOUT divider of 4 */
    CGM.OCDS_SC.R = 0x32000000;   // div by 4, system FMPLL
    CGM.OC_EN.B.EN = 1; 	      // enable CLKOUT signal
    
    /* enable CLK_OUT on ball E20 */
    /* ALT1 - PCR[233] - PA = 0b01 */
    SIU.PCR[233].R = 0x0600;
}

/*******************************************************************************
Function Name : PeriClkGen_init
Engineer      : b08110
Date          : Oct-19-2011
Parameters    : 
Modifies      : 
Returns       : 
Notes         : - Enable auxiliary clock 1-2. from PLL0
Issues        : 
*******************************************************************************/
static void PeriClkGen_init(void) 
{
    CGM.AC0_SC.R = 0x05000000;  /* Select PLL1 for Motor Control Clock  */
    CGM.AC0_DC0.R = 0x80800000;  /* Enable Motor Control clk div by 1 */
    CGM.AC1_SC.R = 0x04000000;  /* Select PLL0 for aux clk 1  */
    CGM.AC1_DC0.R = 0x80000000;  /* Enable aux clk 1 div by 1 */
    CGM.AC2_SC.R = 0x04000000;  /* Select PLL0 for aux clk 2  */
    CGM.AC2_DC0.R = 0x80000000;  /* Enable aux clk 2 div by 1 */
}

/*******************************************************************************
Function Name : DisableWatchdog
Engineer      : b08110
Date          : Jun-23-2010
Parameters    : 
Modifies      : 
Returns       : 
Notes         : - disable SWT functionality
Issues        : 
*******************************************************************************/
void DisableWatchdog(void) 
{
    SWT_0.SR.R = 0x0000c520;     /* Write keys to clear soft lock bit */
    SWT_0.SR.R = 0x0000d928; 
    SWT_0.CR.R = 0x8000010A;     /* Clear watchdog enable (WEN) */
    
    /* e200 Core Watchdog Timer */
    asm
    {
        li  r3, 0
        mtspr   340, r3
    }
}

/*******************************************************************************
Function Name : ClearNCF
Engineer      : b08110
Date          : Oct-19-2011
Parameters    : 
Modifies      : 
Returns       : 
Notes         : Clear NCF 12 only.
Issues        : 
*******************************************************************************/
static void ClearNCF(void)
{
    uint32_t i,b[4];
    for(i=0;i<1;i++)
    {
        FCCU.NCFK.R = FCCU_NCFK_KEY;
        FCCU.NCFS0.R = 0x00001000;  /* clear NCF 12 */
        while(FCCU.CTRL.B.OPS != 0x3)
        {
        
        };              /* wait for the completion of the operation */
        b[i]=FCCU.NCFS0.R;
    }
}

/*******************************************************************************
Function Name : ClearCF
Engineer      : b08110
Date          : Oct-19-2011
Parameters    : 
Modifies      : 
Returns       : 
Notes         : 
Issues        : 
*******************************************************************************/
static void ClearCF(void)
{
    uint32_t i,a[4];
    for(i=0;i<2;i++)
    {
        FCCU.CFK.R = FCCU_CFK_KEY;
   
        FCCU.CFS[i].R = 0xFFFFFFFF;
  
        while(FCCU.CTRL.B.OPS != 0x3)
        {
            
        };      /* wait for the completion of the operation */

        a[i]=FCCU.CFS[i].R;
    }
}

/*******************************************************************************
Function Name : ClearFails
Engineer      : b08110
Date          : Apr-08-2011
Parameters    : 
Modifies      : 
Returns       : 
Notes         : 
Issues        : 
*******************************************************************************/
static void ClearFails(void)
{
    uint16_t reset_register;
    
    if(RGM.FES.B.F_FCCU_SAFE)
    {
        reset_register = RGM.FES.R;
        ME.IMTS.R = 0x00000001;
        ClearCF();
        ClearNCF();
        
        RGM.FES.R = 0x0080; /* clear SAFE mode flag */
        /*RGM.DES.R = 0xFFFF;*/
        
        /* re-enter DRUN */
        ME.MCTL.R = 0x30005AF0; /* Enter DRUN Mode & Key */        
        ME.MCTL.R = 0x3000A50F; /* Enter DRUN Mode & Inverted Key */
    }
}

/*******************************************************************************
Function Name : GPIO_init
Engineer      : b05111
Date          : Nov-10-2014
Parameters    : 
Modifies      : 
Returns       : 
Notes         : 
Issues        : 
*******************************************************************************/
static void GPIO_init(void)
{
    SIU.GPDO[LED1_pad].R = 1;
    SIU.PCR[LED1_pad].B.PA = 0x00; 
    SIU.PCR[LED1_pad].B.OBE = 0x01;    
    
    SIU.GPDO[LED0_pad].R = 1;
    SIU.PCR[LED0_pad].B.PA = 0x00; 
    SIU.PCR[LED0_pad].B.OBE = 0x01;     
}

/*******************************************************************************
Function Name : Delay
Engineer      : b08110
Date          : Apr-08-2011
Parameters    : 
Modifies      : 
Returns       : 
Notes         : 
Issues        : 
*******************************************************************************/
static void Delay(void)
{
    register int32_t cnt = 0;
    
    while(cnt < 1000000)
    {
        cnt++;
    }
}

/*******************************************************************************
Function Name : PIT3_init
Engineer      : b08110
Date          : Feb-16-2012
Parameters    : 
Modifies      : 
Returns       : 
Notes         : - 1 second periodic interrup using the PIT3 module
Issues        : - expecting 45 MHz peripheral clock for the PIT timer
*******************************************************************************/
static void PIT3_init(void)
{
	/* 30: MDIS = 0 to enable clock for PITs. */
    /* 31: FRZ = 1 for Timers stopped in debug mode */
    PIT.PITMCR.R = 0x00000001;    
    PIT.CH[3].LDVAL.R = 45000000 - 1;
    	
    /* clear the TIF flag */
    PIT.CH[3].TFLG.R = 0x00000001;
    	
    /* 30: TIE = 1 for interrupt request enabled */
    /* 31: TEN = 1 for timer active */
    PIT.CH[3].TCTRL.R = 0x00000003;
}

/*******************************************************************************
Function Name : PIT3_isr
Engineer      : b08110
Date          : Mar-12-2011
Parameters    : 
Modifies      : 
Returns       : 
Notes         : - 1 second periodic interrupt using the PIT3 module
Issues        : 
*******************************************************************************/
static void PIT3_isr(void)
{
    /* toggle LED */
    SIU.GPDO[LED0_pad].R ^= 1;

    PIT.CH[3].TFLG.R = 0x00000001;    
}


/*******************************************************************************
Function Name : Terminal_ECHO
Engineer      : b05111
Date          : Apr-19-2012
Parameters    : 
Modifies      : 
Returns       : 
Notes         :                          
Issues        : ReadUARTPoll must be defined in uart.c
*******************************************************************************/
static void Terminal_ECHO(void)
{
    vuint32_t result;
    char ch; 
    
    /* if character is present, read it and re-sent it to terminal */
    result = ReadUARTPoll(&ch);
    if (result==kUARTNoError) 
    {
        switch(ch)
        {
            // enter
            case '\n':
            case '\r':
                printf("\r\n");
                break;
                    
            // other symbols
            default:
                printf("%c",ch); 
                fflush(stdout);
        }           
    } 

}


/*******************************************************************************
* Global functions
*******************************************************************************/

int32_t main(void) 
{
    vint32_t counter = 0;
    
    /* Initialize the MCU */
    HW_init();

    /* Install PIT3_isr into the interrupt vector table */
    INTC_InstallINTCInterruptHandler(PIT3_isr, 127, 1);
    
    /* lower current INTC priority to start INTC interrupts */
    INTC_0.CPR_PRC0.R = 0;
    
    /* 1 sec periodical interrupt */
    PIT3_init();
    
    printf("Hello World\r\n");
    printf("Best Regards, Technical Information Center\r\n");
    printf("You can type..\r\n");      

    /* Loop forever */
    while (1) 
    {
        /* delay */
        while(counter < 100000)
        {
            counter++;
            Terminal_ECHO();

        }  
        counter = 0;
    
        /* toggle LED */
        SIU.GPDO[LED1_pad].R  ^= 1;   
    }
    
}



